from easy_trilateration.graph import *
from easy_trilateration.model import *
from scipy.optimize import least_squares


def rssi_to_distance(rssi, C=17, R=38):
    return 10 ** (-1 * (rssi + C) / R)


def easy_least_squares(crls, guess=(0, 0, 0)):
    result = least_squares(equations, guess, args=[crls])
    xf, yf, rf = result.x

    return Circle(Point(xf, yf), rf), result


def equations(guess, crls: [Circle]):
    eqs = []
    x, y, r = guess
    for circle in crls:
        eqs.append(((x - circle.center.x) ** 2 + (y - circle.center.y) ** 2 - (circle.radius - r) ** 2))
    return eqs


if __name__ == '__main__':
    circles = [Circle(Point(0, 0, ), rssi_to_distance(-72.5958)),
               Circle(Point(16.5, 0), rssi_to_distance(-73.3073)),
               Circle(Point(33, 0), rssi_to_distance(-73.2301)),
               Circle(Point(50, 0), rssi_to_distance(-77.1108)),
               Circle(Point(0, 50), rssi_to_distance(-77.1108)),
               Circle(Point(16.5, 50), rssi_to_distance(-73.3082)),
               Circle(Point(33, 50), rssi_to_distance(-73.231)),
               Circle(Point(50, 50), rssi_to_distance(-77.1113)),
               Circle(Point(0, 25), rssi_to_distance(-72.5948)),
               Circle(Point(50, 25), rssi_to_distance(-72.5948))]
    result_cirl, meta = easy_least_squares(circles)
    create_circle(result_cirl, target=True)
    draw([circles[0], circles[1], circles[2], circles[3], circles[4], circles[5], circles[6], circles[7],
          circles[8], circles[9]])



from math import sqrt, hypot


class Point:
    x: float
    y: float

    def __init__(self, x_init: float, y_init: float):
        self.x = x_init
        self.y = y_init

    def __repr__(self):
        return "".join(["Point(", str(self.x), ",", str(self.y), ")"])

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Point):
            return (self.x == other.x) & (self.y == other.y)
        return False

    def __hash__(self):
        return hash(str(self))

    def shift(self, x, y):
        self.x += x
        self.y += y

    def sum(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def subtract(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def multiply(self, factor):
        return Point(self.x * factor, self.y * factor)

    def distance(self, b):
        return hypot(self.x - b.x, self.y - b.y)


class Circle:
    center: Point
    radius: float

    def __init__(self, point_init: Point, r_init: float):
        self.center = point_init
        self.radius = r_init

    def __repr__(self):
        return "".join(["Circle(", str(self.center.x), ", ", str(self.center.y), ", ", str(self.radius), ")"])

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Circle):
            return (self.center.x == other.center.x) & (self.center.y == other.center.y) & (self.radius == other.radius)
        return False

    def shift(self, x, y):
        self.center.shift(x, y)

    def shift_radius(self, r):
        self.radius += r

    def has_inside(self, point: Point):
        return self.center.distance(point) <= self.radius

    def intersects(self, other):
        if not isinstance(other, Circle):  # if not a circle
            return False
        if self.center.distance(other.center) > (self.radius + other.radius):
            return False  # They dont intersect
        elif self.center.distance(other.center) < abs(self.radius - other.radius):
            return False  # One is inside the other
        elif (self.center.distance(other.center) == 0) and (self.radius == other.radius):
            return False  # They are exactly the same circle
        else:  # if distance(a.center,b.center)<=a.r+b.r
            return True

    def distance_to_intersection(self, other):
        if not isinstance(other, Circle):  # if not a circle
            return None
        return self.center.distance(other.center) - (self.radius + other.radius)

    def get_intersection_points(self, other, precision: int = None):
        if not self.intersects(other):
            return {}

        distance_between_centers = self.center.distance(other.center)
        if precision is not None:
            distance_between_centers = round(distance_between_centers, precision)

        distance_between_x = other.center.x - self.center.x
        distance_between_y = other.center.y - self.center.y
        distance_center_to_middle = (self.radius ** 2 - other.radius ** 2 + distance_between_centers ** 2) / (
                2 * distance_between_centers)
        distance_between_intersection_points = sqrt(self.radius ** 2 - distance_center_to_middle ** 2)
        intersection_center_x = self.center.x + (
                distance_center_to_middle * distance_between_x / distance_between_centers)
        intersection_center_y = self.center.y + (
                distance_center_to_middle * distance_between_y / distance_between_centers)

        intersection_point_a_x = intersection_center_x + (
                distance_between_intersection_points * distance_between_y) / distance_between_centers
        intersection_point_a_y = intersection_center_y - (
                distance_between_intersection_points * distance_between_x) / distance_between_centers
        if precision is not None:
            intersection_point_a_x = round(intersection_point_a_x, precision)
            intersection_point_a_y = round(intersection_point_a_y, precision)
        intersection_point_a = Point(intersection_point_a_x, intersection_point_a_y)

        intersection_point_b_x = intersection_center_x - (
                distance_between_intersection_points * distance_between_y) / distance_between_centers
        intersection_point_b_y = intersection_center_y + (
                distance_between_intersection_points * distance_between_x) / distance_between_centers
        if precision is not None:
            intersection_point_b_x = round(intersection_point_b_x, precision)
            intersection_point_b_y = round(intersection_point_b_y, precision)
        intersection_point_b = Point(intersection_point_b_x, intersection_point_b_y)

        return {intersection_point_a, intersection_point_b}


def find_intersections(circles: [Circle]):
    ret = set()
    for item in circles:
        for item2 in circles:
            if item.intersects(item2):
                for item3 in item.get_intersection_points(item2):
                    print("intersects on " + str(item3))
                    ret.add(item3)
    return ret



from setuptools import setup
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='easy_trilateration',
    version='0.1.2',
    description='Easy Least-Squares trilateration using scipy',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/agusalex/Least-Squares-Trilateration',
    author='Agustin Alexander',
    author_email='agustin.c.alexander@gmail.com',
    license='MIT',
    packages=['easy_trilateration'],
    install_requires=['matplotlib',
                      'scipy'
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
    ],
)



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from random import randint

from easy_trilateration import model


def live_plotter(x_vec, y1_data, line1, identifier='', pause_time=0.1):
    if not line1:
        # this is the call to matplotlib that allows dynamic plotting
        plt.ion()
        fig = plt.figure(figsize=(13, 6))
        ax = fig.add_subplot(111)
        # create a variable for the line so we can later update it
        line1, = ax.plot(x_vec, y1_data, '-o', alpha=0.8)
        # update plot label/title
        plt.ylabel('Y Label')
        plt.title('Title: {}'.format(identifier))
        plt.show()

    # after the figure, axis, and line are created, we only need to update the y-data
    line1.set_ydata(y1_data)
    # adjust limits if new data goes beyond bounds
    if np.min(y1_data) <= line1.axes.get_ylim()[0] or np.max(y1_data) >= line1.axes.get_ylim()[1]:
        plt.ylim([np.min(y1_data) - np.std(y1_data), np.max(y1_data) + np.std(y1_data)])
    # this pauses the data so the figure/axis can catch up - the amount of pause can be altered above
    plt.pause(pause_time)

    # return line so we can update it again in the next iteration
    return line1


def demolive():
    plt.style.use('ggplot')
    size = 100
    x_vec = np.linspace(0, 1, size + 1)[0:-1]
    y_vec = np.random.randn(len(x_vec))
    line1 = []
    while True:
        rand_val = np.random.randn(1)
        y_vec[-1] = rand_val
        line1 = live_plotter(x_vec, y_vec, line1)
        y_vec = np.append(y_vec[1:], 0.0)


def create_circle(circle: model.Circle, target=False):
    color = matplotlib.cm.jet(randint(50, 100))
    if target:
        color = matplotlib.cm.jet(1000)
    add_shape(plt.Circle((circle.center.x, circle.center.y), color=color, fill=False, zorder=1, radius=circle.radius,
                         alpha=0.8))
    plt.scatter(circle.center.x, circle.center.y, color=color, s=100, zorder=2)


def create_point(point: model.Point, color=matplotlib.cm.jet(randint(0, 00))):
    plt.scatter(point.x, point.y, color=color, s=100, zorder=2)


def add_shape(patch):
    ax = plt.gca()
    ax.add_patch(patch)
    plt.axis('scaled')


def draw(drawlist):
    for item in drawlist:
        if isinstance(item, model.Circle):
            create_circle(item)
        if isinstance(item, model.Point):
            create_point(item)
    plt.show()



